""" Command line interface for FMPy """

from __future__ import print_function


def main():

    import argparse
    import textwrap

    description = """\
    Validate and simulate Functional Mock-up Units (FMUs)

    Get information about an FMU:
       
        fmpy info Rectifier.fmu
     
    Simulate an FMU:
     
        fmpy simulate Rectifier.fmu --show-plot
        
    Compile a source code FMU:
    
        fmpy compile Rectifier.fmu
        
    Create a Jupyter Notebook
    
        fmpy create-jupyter-notebook Rectifier.fmu
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=textwrap.dedent(description))

    parser.add_argument('command', choices=['info', 'validate', 'simulate', 'compile', 'add-cswrapper', 'add-remoting',
                                            'create-cmake-project', 'create-jupyter-notebook'],
                        help="Command to execute")
    parser.add_argument('fmu_filename', help="filename of the FMU")

    parser.add_argument('--validate', action='store_true', help="validate the FMU")
    parser.add_argument('--start-time', type=float, help="start time for the simulation")
    parser.add_argument('--stop-time', type=float, help="stop time for the simulation")
    parser.add_argument('--solver', choices=['Euler', 'CVode'], default='CVode', help="solver to use for Model Exchange")
    parser.add_argument('--step-size', type=float, help="step size for fixed-step solvers")
    parser.add_argument('--relative-tolerance', type=float, help="relative tolerance for the 'CVode' solver and FMI 2.0 co-simulation FMUs")
    parser.add_argument('--dont-record-events', action='store_true', help="dont't record outputs at events (model exchange only)")
    parser.add_argument('--start-values', nargs='+', help="name-value pairs of start values")
    parser.add_argument('--apply-default-start-values', action='store_true', help="apply the start values from the model description")
    parser.add_argument('--output-interval', type=float, help="interval for sampling the output")
    parser.add_argument('--input-file', help="CSV file to use as input")
    parser.add_argument('--output-variables', nargs='+', help="Variables to record")
    parser.add_argument('--output-file', help="CSV to store the results")
    parser.add_argument('--timeout', type=float, help="max. time to wait for the simulation to finish")
    parser.add_argument('--debug-logging', action='store_true', help="enable the FMU's debug logging")
    parser.add_argument('--visible', action='store_true', help="enable interactive mode")
    parser.add_argument('--fmi-logging', action='store_true', help="enable FMI logging")
    parser.add_argument('--show-plot', action='store_true', help="plot the results")
    parser.add_argument('--cmake-project-dir', help="Directory for the CMake project")

    args = parser.parse_args()

    if args.command == 'info':

        from fmpy import dump
        dump(args.fmu_filename)

    elif args.command == 'validate':

        import sys
        from fmpy.validation import validate_fmu

        problems = validate_fmu(args.fmu_filename)

        if len(problems) == 0:
            print('No problems found.')
        else:
            print('%d problems were found:' % len(problems))
            for message in problems:
                print()
                print(message)

        sys.exit(len(problems))

    elif args.command == 'compile':

        from fmpy.util import compile_platform_binary
        compile_platform_binary(args.fmu_filename)

    elif args.command == 'add-cswrapper':

        from fmpy.cswrapper import add_cswrapper
        add_cswrapper(args.fmu_filename)

    elif args.command == 'add-remoting':

        from fmpy.util import add_remoting
        add_remoting(args.fmu_filename)

    elif args.command == 'create-cmake-project':

        import os
        from fmpy.util import create_cmake_project

        project_dir = args.cmake_project_dir

        if project_dir is None:
            project_dir = os.path.basename(args.fmu_filename)
            project_dir, _ = os.path.splitext(project_dir)
            print("Creating CMake project in %s" % os.path.abspath(project_dir))

        create_cmake_project(args.fmu_filename, project_dir)

    elif args.command == 'create-jupyter-notebook':

        from fmpy.util import create_jupyter_notebook

        create_jupyter_notebook(args.fmu_filename)

    elif args.command == 'simulate':

        from fmpy import simulate_fmu
        from fmpy.util import read_csv, write_csv, plot_result

        if args.start_values:
            if len(args.start_values) % 2 != 0:
                raise Exception("Start values must be name-value pairs.")
            start_values = {k: v for k, v in zip(args.start_values[::2], args.start_values[1::2])}
        else:
            start_values = {}

        input = read_csv(args.input_file) if args.input_file else None

        if args.fmi_logging:
            fmi_call_logger = lambda s: print('[FMI] ' + s)
        else:
            fmi_call_logger = None

        result = simulate_fmu(args.fmu_filename,
                              validate=args.validate,
                              start_time=args.start_time,
                              stop_time=args.stop_time,
                              solver=args.solver,
                              step_size=args.step_size,
                              relative_tolerance=args.relative_tolerance,
                              output_interval=args.output_interval,
                              record_events=not args.dont_record_events,
                              fmi_type=None,
                              start_values=start_values,
                              apply_default_start_values=args.apply_default_start_values,
                              input=input,
                              output=args.output_variables,
                              timeout=args.timeout,
                              debug_logging=args.debug_logging,
                              visible=args.visible,
                              fmi_call_logger=fmi_call_logger)

        if args.output_file:
            write_csv(filename=args.output_file, result=result)

        if args.show_plot:
            plot_result(result=result, window_title=args.fmu_filename)



""" Entry point for the graphical user interface """

if __name__ == '__main__':

    import os
    import sys
    import ctypes
    import platform
    from PyQt5 import QtCore
    from PyQt5.QtWidgets import QApplication
    from fmpy.gui.MainWindow import MainWindow

    if os.name == 'nt' and int(platform.release()) >= 8:
        ctypes.windll.shcore.SetProcessDpiAwareness(True)

    QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()

    for i, v in enumerate(sys.argv[1:]):
        if i > 0:
            window = MainWindow()
            window.show()
        window.load(v)

    sys.exit(app.exec_())



from lxml import etree
import zipfile
import os

# namespaces for XML parsing
ns = {
    'ssc': 'http://ssp-standard.org/SSP1/SystemStructureCommon',
    'ssd': 'http://ssp-standard.org/SSP1/SystemStructureDescription',
    'ssm': 'http://ssp-standard.org/SSP1/SystemStructureParameterMapping',
    'ssv': 'http://ssp-standard.org/SSP1/SystemStructureParameterValues',
    'ssb': 'http://ssp-standard.org/SSP1/SystemStructureSignalDictionary',
}


# common
class BaseElement(object):

    def __init__(self, **kwargs):
        super(BaseElement, self).__init__()
        self.id = kwargs.get('id', None)
        self.description = kwargs.get('description', None)


class TopLevelMetaData(object):

    def __init__(self, **kwargs):
        super(TopLevelMetaData, self).__init__()
        self.author = kwargs.get('author', None)
        self.fileversion = kwargs.get('fileversion', None)
        self.copyright = kwargs.get('copyright', None)
        self.license = kwargs.get('license', None)
        self.generationTool = kwargs.get('generationTool', None)
        self.generationDateAndTime = kwargs.get('generationDateAndTime', None)


class LinearTransformation(object):

    def __init__(self, factor=1.0, offset=0.0):
        super(LinearTransformation, self).__init__()
        self.factor = factor
        self.offset = offset

    def __repr__(self):
        return "LinearTransformation (factor: %s, offset: %s)" % (self.factor, self.offset)


class BooleanMappingTransformation(object):

    def __init__(self):
        super(BooleanMappingTransformation, self).__init__()
        self.entries = {}

    def __repr__(self):
        return "BooleanMappingTransformation (entries: %s)" % (self.entries,)


class IntegerMappingTransformation(object):

    def __init__(self):
        super(IntegerMappingTransformation, self).__init__()
        self.entries = {}

    def __repr__(self):
        return "IntegerMappingTransformation (entries: %s)" % (self.entries,)


class EnumerationMappingTransformation(object):

    def __init__(self):
        super(EnumerationMappingTransformation, self).__init__()
        self.entries = {}

    def __repr__(self):
        return "EnumerationMappingTransformation (entries: %s)" % (self.entries,)


# system structure description
class SystemStructureDescription(BaseElement, TopLevelMetaData):

    def __init__(self, **kwargs):
        super(SystemStructureDescription, self).__init__(**kwargs)
        self.system = kwargs.get('system')
        self.enumerations = kwargs.get('enumerations', [])
        self.units = kwargs.get('units', [])
        self.defaultExperiment = kwargs.get('defaultExperiment')
        self.annotations = kwargs.get('annotations', [])
        self.version = kwargs.get('version')
        self.name = kwargs.get('name')

    def __repr__(self):
        return "SystemStructureDescription (name: %s)" % self.name


class Element(BaseElement):

    def __init__(self, **kwargs):
        super(Element, self).__init__(**kwargs)
        self.connectors = kwargs.get('connectors', [])
        self.parameterBindings = kwargs.get('parameterBindings', [])
        self.name = kwargs.get('name')


class Component(Element):

    def __init__(self, **kwargs):
        super(Component, self).__init__(**kwargs)
        self.type = kwargs.get('type')
        self.source = kwargs.get('source')
        self.implementation = kwargs.get('implementation')

    def __repr__(self):
        return "Component (name: %s, source: %s, type: %s)" % (self.name, self.source, self.type)


class System(Element):

    def __init__(self, **kwargs):
        super(System, self).__init__(**kwargs)
        self.elements = kwargs.get('elements', [])
        self.connections = kwargs.get('connections', [])
        self.signalDictionaries = kwargs.get('signalDictionaries', [])
        self.systemGeometry = kwargs.get('systemGeometry')
        self.graphicalElements = kwargs.get('graphicalElements', [])
        self.simulationInformation = kwargs.get('simulationInformation')
        self.annotations = kwargs.get('annotations', [])

    def __repr__(self):
        return "System (name: %s, description: %s)" % (self.name, self.description)


class Connector(BaseElement):

    def __init__(self, **kwargs):
        super(Connector, self).__init__(**kwargs)
        self.name = kwargs.get('name')
        self.kind = kwargs.get('kind')

    def __repr__(self):
        return "Connector (name: %s, kind: %s)" % (self.name, self.kind)


class Connection(BaseElement):

    def __init__(self, **kwargs):
        super(Connection, self).__init__(**kwargs)
        self.startElement = kwargs.get('startElement')
        self.startConnector = kwargs.get('startConnector')
        self.endElement = kwargs.get('endElement')
        self.endConnector = kwargs.get('endConnector')

    def __repr__(self):
        return "Connection (startElement: %s, startConnector: %s, endElement: %s, endConnector: %s)" % (self.startElement, self.startConnector, self.endElement, self.endConnector)


class ParameterBinding(BaseElement):

    def __init__(self, **kwargs):
        super(ParameterBinding, self).__init__(**kwargs)
        self.parameterValues = kwargs.get('parameterValues', [])
        self.parameterMapping = kwargs.get('parameterMapping', [])
        self.annotations = kwargs.get('annotations', [])
        self.type = kwargs.get('type')
        self.source = kwargs.get('source')
        self.sourceBase = kwargs.get('sourceBase')
        self.prefix = kwargs.get('prefix')

    def __repr__(self):
        return "ParameterBinding (prefix: %s, source: %s, type: %s)" % (self.prefix, self.source, self.type)


class DefaultExperiment(object):

    def __init__(self):
        super(DefaultExperiment, self).__init__()
        self.annotations = []
        self.startTime = None
        self.stopTime = None


class SimulationInformation(object):

    def __init__(self, **kwargs):
        super(SimulationInformation, self).__init__()
        self.fixedStepSolver = kwargs.get('fixedStepSolver')
        self.variableStepSolver = kwargs.get('variableStepSolver')
        self.fixedStepMaster = kwargs.get('fixedStepMaster')
        self.variableStepMaster = kwargs.get('variableStepMaster')
        self.annotations = kwargs.get('annotations', [])


class FixedStepSolver(object):

    def __init__(self, **kwargs):
        super(FixedStepSolver, self).__init__()
        self.description = kwargs.get('description')
        self.stepSize = kwargs.get('stepSize')


class VariableStepSolver(object):

    def __init__(self, **kwargs):
        super(VariableStepSolver, self).__init__()
        self.description = kwargs.get('description')
        self.absoluteTolerance = kwargs.get('absoluteTolerance')
        self.relativeTolerance = kwargs.get('relativeTolerance')
        self.minimumStepSize = kwargs.get('minimumStepSize')
        self.maximumStepSize = kwargs.get('maximumStepSize')
        self.initialStepSize = kwargs.get('initialStepSize')


class FixedStepMaster(object):

    def __init__(self, **kwargs):
        super(FixedStepMaster, self).__init__()
        self.description = kwargs.get('description')
        self.stepSize = kwargs.get('stepSize')


class VariableStepMaster(object):

    def __init__(self, **kwargs):
        super(VariableStepMaster, self).__init__()
        self.description = kwargs.get('description')
        self.minimumStepSize = kwargs.get('minimumStepSize')
        self.maximumStepSize = kwargs.get('maximumStepSize')
        self.initialStepSize = kwargs.get('initialStepSize')


# parameter mapping
class ParameterMapping(BaseElement, TopLevelMetaData):

    def __init__(self, **kwargs):
        super(ParameterMapping, self).__init__(**kwargs)
        self.mappingEntries = kwargs.get('mappingEntries', [])
        self.annotations = kwargs.get('annotations', [])
        self.version = kwargs.get('version')


class MappingEntry(BaseElement):

    def __init__(self, **kwargs):
        super(MappingEntry, self).__init__(**kwargs)
        self.transformation = kwargs.get('transformation')
        self.source = kwargs.get('source')
        self.target = kwargs.get('target')
        self.suppressUnitConversion = kwargs.get('suppressUnitConversion')


class SignalDictionaryReference(object):

    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.dictionary = kwargs.get('dictionary')
        self.connectors = kwargs.get('connectors', [])

    def __repr__(self):
        return "SignalDictionaryReference (name: %s, dictionary: %s)" % (self.name, self.dictionary)


# signal dictionary
class SignalDictionary(BaseElement, TopLevelMetaData):

    def __init__(self, **kwargs):
        super(SignalDictionary, self).__init__(**kwargs)
        self.entries = kwargs.get('entries', [])
        self.enumerations = kwargs.get('enumerations', [])
        self.units = kwargs.get('units', [])
        self.annotations = kwargs.get('annotations', [])
        self.version = kwargs.get('version')
        self.type = kwargs.get('type')
        self.source = kwargs.get('source')
        self.name = kwargs.get('name')

    def __repr__(self):
        return "SignalDictionary"


class DictionaryEntry(BaseElement):

    def __init__(self, **kwargs):
        super(DictionaryEntry, self).__init__(**kwargs)
        self.name = kwargs.get('name')
        self.type = kwargs.get('type')
        self.unit = kwargs.get('unit')

    def __repr__(self):
        return "DictionaryEntry (name: %s, type: %s, unit=%s)" % (self.name, self.type, self.unit)


class Unit(object):

    def __init__(self, **kwargs):
        super(Unit, self).__init__()
        self.name = kwargs.get('name')
        self.kg = int(kwargs.get('kg', 0))
        self.m = int(kwargs.get('m', 0))
        self.s = int(kwargs.get('s', 0))
        self.A = int(kwargs.get('A', 0))
        self.K = int(kwargs.get('K', 0))
        self.mol = int(kwargs.get('mol', 0))
        self.cd = int(kwargs.get('cd', 0))
        self.rad = int(kwargs.get('rad', 0))
        self.factor = float(kwargs.get('factor', 1.0))
        self.offset = float(kwargs.get('offset', 0.0))

    def __repr__(self):
        return "Unit (name: %s, kg: %d, m: %d, s: %d, A: %d, K: %d, mol: %d, cd: %d, rad: %d, factor: %g, offset: %g)" % (self.name, self.kg, self.m, self.s, self.A, self.K, self.mol, self.cd, self.rad, self.factor, self.offset)


class ParameterSet(BaseElement, TopLevelMetaData):

    def __init__(self, **kwargs):
        super(ParameterSet, self).__init__(**kwargs)
        self.parameters = kwargs.get('parameters', [])
        self.enumerations = kwargs.get('enumerations', [])
        self.units = kwargs.get('units', [])
        self.annotations = kwargs.get('annotations', [])
        self.version = kwargs.get('version')
        self.name = kwargs.get('name')

    def __repr__(self):
        return "ParameterSet (name: %s)" % self.name


class Parameter(BaseElement):

    def __init__(self, **kwargs):
        super(Parameter, self).__init__()
        self.name = kwargs.get('name')
        self.type = kwargs.get('type')
        self.value = kwargs.get('value')
        self.unit = kwargs.get('unit')

    def __repr__(self):
        return "Parameter (name: %s, type: %s, value: %s, unit: %s)" % (self.name, self.type, self.value, self.unit)


def validate_tree(root, schema_file):

    module_dir, _ = os.path.split(__file__)

    schema = etree.XMLSchema(file=os.path.join(module_dir, 'schema', schema_file))

    if not schema.validate(root):
        message = "Failed to validate SystemStructure.ssd:"
        for entry in schema.error_log:
            message += "\n%s (line %d, column %d): %s" % (entry.level_name, entry.line, entry.column, entry.message)
        raise Exception(message)


def read_ssv(filename, resource=None, validate=True):

    if resource is None:
        tree = etree.parse(filename)
    else:
        # load parameter set
        with zipfile.ZipFile(filename, 'r') as zf:
            xml = zf.open(resource)
            tree = etree.parse(xml)

    root = tree.getroot()

    if validate:
        validate_tree(root, 'SystemStructureParameterValues.xsd')

    return _get_parameter_set(root)


def _get_transformation(element):

    for t in element.findall('ssc:LinearTransformation', namespaces=ns):
        return LinearTransformation(**t.attrib)

    # TODO:
    # for type in [BooleanMappingTransformation, IntegerMappingTransformation, EnumerationMappingTransformation]:
    #     pass

    return None


def _get_parameter_set(element):

    parameter_set = ParameterSet(name=element.get('name'))

    for p in element.findall('ssv:Parameters/ssv:Parameter', namespaces=ns):

        parameter = Parameter(name=p.get('name'))
        real = p.find('ssv:Real', namespaces=ns)

        if real is not None:
            parameter.type = 'Real'
            parameter.value = real.get('value')
            parameter.unit = real.get('unit')
        else:
            for type in ['Integer', 'Boolean', 'String', 'Enumeration', 'Binary']:
                entry = p.find('ssv:' + type, namespaces=ns)
                if entry is not None:
                    parameter.type = type
                    parameter.value = entry.get('value')
                    break

        parameter_set.parameters.append(parameter)

    return parameter_set


def read_ssm(filename, resource=None, validate=True):

    if resource is None:
        tree = etree.parse(filename)
    else:
        # load parameter set
        with zipfile.ZipFile(filename, 'r') as zf:
            xml = zf.open(resource)
            tree = etree.parse(xml)

    root = tree.getroot()

    if validate:
        validate_tree(root, 'SystemStructureParameterMapping.xsd')

    parameter_mapping = ParameterMapping()

    for m in root.findall('ssm:MappingEntry', namespaces=ns):
        mapping_entry = MappingEntry(**m.attrib)
        mapping_entry.transformation = _get_transformation(m)
        parameter_mapping.mappingEntries.append(mapping_entry)

    return parameter_mapping


def _handle_signal_dictionary(element):

    dictionary = SignalDictionary(**element.attrib)

    for e in element.findall('ssb:DictionaryEntry', namespaces=ns):

        entry = DictionaryEntry(name=e.get('name'))

        real = e.find('ssc:Real', namespaces=ns)
        if real is not None:
            entry.type = 'Real'
            entry.unit = real.get('unit')
        else:
            for type in ['Integer', 'Boolean', 'String', 'Enumeration', 'Binary']:
                entry = e.find('ssc:' + type, namespaces=ns)
                if entry is not None:
                    entry.type = type
                    break

        dictionary.entries.append(entry)

    return dictionary


def get_connectors(element):

    connectors = []

    for c in element.findall('ssd:Connectors/ssd:Connector', namespaces=ns):
        connector = Connector(name=c.get('name'), kind=c.get('kind'))
        connectors.append(connector)

    return connectors


def _handle_element(object, element, filename):

    for b in element.findall('ssd:ParameterBindings/ssd:ParameterBinding', namespaces=ns):

        parameter_binding = ParameterBinding(**b.attrib)

        if parameter_binding.source is not None:
            parameter_set = read_ssv(filename, resource=parameter_binding.source)
            parameter_binding.parameterValues.append(parameter_set)

        for s in b.findall('ssd:ParameterValues/ssv:ParameterSet', namespaces=ns):
            parameter_set = _get_parameter_set(s)
            parameter_binding.parameterValues.append(parameter_set)

        for m in b.findall('ssd:ParameterMapping', namespaces=ns):

            source = m.get('source')

            if source is None:
                parameter_mapping = ParameterMapping()
            else:
                parameter_mapping = read_ssm(filename, resource=source)

            parameter_binding.parameterMapping.append(parameter_mapping)

        object.parameterBindings.append(parameter_binding)


def handle_system(system, filename):
    """
    Parameters:
        system      ...
        filename    filename of the SSP file

    Returns:
        A System object
    """

    system_obj = System()

    system_obj.name = system.get('name')
    system_obj.description = system.get('description')
    _handle_element(system_obj, system, filename)

    # for b in system.findall('ssd:ParameterBindings/ssd:ParameterBinding', namespaces=ns):
    #     system_obj.parameterBindings.append(
    #         ParameterBinding(prefix=b.get('prefix'), source=b.get('source'), type=b.get('type')))

    system_obj.connectors = get_connectors(system)

    # Components
    for c in system.findall('ssd:Elements/ssd:Component', namespaces=ns):
        component = Component(name=c.get('name'), source=c.get('source'), type=c.get('type'))
        _handle_element(component, c, filename)
        component.connectors = get_connectors(c)
        system_obj.elements.append(component)

    # SignalDictionaryReferences
    for r in system.findall('ssd:Elements/ssd:SignalDictionaryReference', namespaces=ns):
        reference = SignalDictionaryReference(name=r.get('name'), dictionary=r.get('dictionary'))
        reference.connectors = get_connectors(r)
        system_obj.elements.append(reference)

    # Connections
    for c in system.findall('ssd:Connections/ssd:Connection', namespaces=ns):
        connection = Connection(startElement=c.get('startElement'),
                                startConnector=c.get('startConnector'),
                                endElement=c.get('endElement'),
                                endConnector=c.get('endConnector'))
        # LinearTransformation
        t = c.find('ssd:LinearTransformation', namespaces=ns)
        if t is not None:
            connection.linearTransformation = LinearTransformation(factor=float(t.get('factor', 1.0)),
                                                                   offset=float(t.get('offset', 0.0)))

        system_obj.connections.append(connection)

    # SignalDictionaries
    for d in system.findall('ssd:SignalDictionaries/ssd:SignalDictionary', namespaces=ns):

        source = d.get('source')

        if source is not None:
            # referenced
            with zipfile.ZipFile(filename, 'r') as zf:
                xml = zf.open(source)
                tree = etree.parse(xml)
                sd = tree.getroot()
        else:
            # inline
            sd = d.find('ssb:SignalDictionary', namespaces=ns)

        signal_dictionary = _handle_signal_dictionary(sd)

        signal_dictionary.name = d.get('name')
        signal_dictionary.source = d.get('source')
        signal_dictionary.type = d.get('type')

        system_obj.signalDictionaries.append(signal_dictionary)

    # Systems
    for system in system.findall('ssd:Elements/ssd:System', namespaces=ns):
        child_system_obj = handle_system(system, filename=filename)
        system_obj.elements.append(child_system_obj)

    return system_obj


def read_ssd(filename, validate=True):

    with zipfile.ZipFile(filename, 'r') as zf:
        xml = zf.open('SystemStructure.ssd')
        tree = etree.parse(xml)

    root = tree.getroot()

    if validate:
        validate_tree(root, 'SystemStructureDescription.xsd')

    ssd = SystemStructureDescription()

    ssd.version = root.get('version')
    ssd.name = root.get('name')

    # Units
    for u in root.findall('ssd:Units/ssc:Unit', namespaces=ns):
        attr = dict(u.attrib)
        bu = u.find('ssc:BaseUnit', namespaces=ns)
        attr.update(bu.attrib)
        ssd.units.append(Unit(**attr))
    ssd.system = System()

    system = root.find('ssd:System', namespaces=ns)

    ssd.system = handle_system(system, filename=filename)

    # add parent elements
    add_tree_info(ssd.system)
    ssd.system.parent = None

    return ssd


def add_tree_info(element):

    if isinstance(element, System):
        for child in element.elements:
            child.parent = element
            add_tree_info(child)

    for connector in element.connectors:
        connector.parent = element


def build_path(object):

    path = object.name

    while object.parent is not None:
        object = object.parent
        path = object.name + '.' + path

    return path


def find_connectors(element):
    """ Return a list of all connectors in element including (sub-systems)"""

    connectors = []
    connectors += element.connectors

    if isinstance(element, System):
        for child in element.elements:
            connectors += find_connectors(child)

    return connectors


def find_components(system):
    """ Return a list of all components in system including (sub-systems)"""

    components = []

    for element in system.elements:
        if isinstance(element, Component):
            components.append(element)
        elif isinstance(element, System):
            components += find_components(element)

    return components


def get_connections(system, connectors=None):
    """ Create a list of (start_connector, end_connector) from all connections in the system """

    if connectors is None:
        connectors = {}  # path -> object
        for connector in find_connectors(system):
            connectors[build_path(connector)] = connector

    cons = []

    # connections to the outside
    for connector in system.connectors:

        if connector.kind == 'output':

            # find the connection
            for connection in system.connections:
                if connection.endElement is None and connection.endConnector == connector.name:
                    start_p = build_path(system) + '.' + connection.startElement + '.' + connection.startConnector
                    end_p = build_path(connector)
                    break
                elif connection.startElement is None and connection.startConnector == connector.name:
                    start_p = build_path(connector)
                    end_p = build_path(system) + '.' + connection.startElement + '.' + connection.startConnector
                    break

            cons.append((connectors[start_p], connectors[end_p]))

    # internal connections
    for element in system.elements:

        for connector in element.connectors:

            if connector.kind == 'input':
                end_p = build_path(element) + '.' + connector.name

                # find the connection
                for connection in system.connections:
                    if connection.endElement == element.name and connection.endConnector == connector.name:
                        start_p = build_path(system)
                        if connection.startElement is not None:
                            start_p += '.' + connection.startElement
                        start_p += '.' + connection.startConnector
                        break
                    elif connection.startElement == element.name and connection.startConnector == connector.name:
                        start_p = build_path(system)
                        if connection.endElement is not None:
                            start_p += '.' + connection.endElement
                        start_p += '.' + connection.endConnector
                        break

                cons.append((connectors[start_p], connectors[end_p]))

    # continue with the child systems
    for element in system.elements:
        if isinstance(element, System):
            cons += get_connections(element, connectors=connectors)

    return cons



from tempfile import mkdtemp


def create_fmu_container(configuration, output_filename):
    """ Create an FMU from nested FMUs (experimental)

        see tests/test_fmu_container.py for an example
    """

    import os
    import shutil
    import fmpy
    from fmpy import read_model_description, extract
    import msgpack
    from datetime import datetime
    import pytz

    base_filename, _ = os.path.splitext(output_filename)
    model_name = os.path.basename(base_filename)

    unzipdir = mkdtemp()

    basedir = os.path.dirname(__file__)

    for directory in ['binaries', 'documentation', 'sources']:
        shutil.copytree(os.path.join(basedir, directory), os.path.join(unzipdir, directory))

    os.mkdir(os.path.join(unzipdir, 'resources'))

    data = {
        'components': [],
        'variables': [],
        'connections': []
    }

    l = []

    component_map = {}
    vi = 0  # variable index

    l.append('<?xml version="1.0" encoding="UTF-8"?>')
    l.append('<fmiModelDescription')
    l.append('  fmiVersion="2.0"')
    l.append('  modelName="%s"' % model_name)
    l.append('  guid=""')
    if 'description' in configuration:
        l.append('  description="%s"' % configuration['description'])
    l.append('  generationTool="FMPy %s FMU Container"' % fmpy.__version__)
    l.append('  generationDateAndTime="%s">' % datetime.now(pytz.utc).isoformat())
    l.append('')
    l.append('  <CoSimulation modelIdentifier="FMUContainer">')
    l.append('    <SourceFiles>')
    l.append('      <File name="FMUContainer.c"/>')
    l.append('      <File name="mpack.c"/>')
    l.append('    </SourceFiles>')
    l.append('  </CoSimulation>')
    l.append('')
    l.append('  <ModelVariables>')
    for i, component in enumerate(configuration['components']):
        model_description = read_model_description(component['filename'])
        model_identifier = model_description.coSimulation.modelIdentifier
        extract(component['filename'], os.path.join(unzipdir, 'resources', model_identifier))
        variables = dict((v.name, v) for v in model_description.modelVariables)
        component_map[component['name']] = (i, variables)
        data['components'].append({
            'name': component['name'],
            'guid': model_description.guid,
            'modelIdentifier': model_identifier,
        })
        for name in component['variables']:
            v = variables[name]
            data['variables'].append({'component': i, 'valueReference': v.valueReference})
            name = component['name'] + '.' + v.name
            description = v.description
            if name in configuration['variables']:
                mapping = configuration['variables'][name]
                if 'name' in mapping:
                    name = mapping['name']
                if 'description' in mapping:
                    description = mapping['description']
            description = ' description="%s"' % description if description else ''
            l.append('    <ScalarVariable name="%s" valueReference="%d" causality="%s" variability="%s"%s>' % (name, vi, v.causality, v.variability, description))
            l.append('      <%s%s/>' % (v.type, ' start="%s"' % v.start if v.start else ''))
            l.append('    </ScalarVariable>')
            vi += 1
    l.append('  </ModelVariables>')
    l.append('')
    l.append('  <ModelStructure/>')
    l.append('')
    l.append('</fmiModelDescription>')

    for sc, sv, ec, ev in configuration['connections']:
        data['connections'].append({
            'type': component_map[sc][1][sv].type,
            'startComponent': component_map[sc][0],
            'endComponent': component_map[ec][0],
            'startValueReference': component_map[sc][1][sv].valueReference,
            'endValueReference': component_map[ec][1][ev].valueReference,
        })

    with open(os.path.join(unzipdir, 'modelDescription.xml'), 'w') as f:
        f.write('\n'.join(l) + '\n')

    with open(os.path.join(unzipdir, 'resources', 'config.mp'), 'wb') as f:
        packed = msgpack.packb(data)
        f.write(packed)

    shutil.make_archive(base_filename, 'zip', unzipdir)

    if os.path.isfile(output_filename):
        os.remove(output_filename)

    os.rename(base_filename + '.zip', output_filename)

    shutil.rmtree(unzipdir, ignore_errors=True)



from ctypes import *

# /* -----------------------------------------------------------------
#  * Programmer(s): Scott Cohen, Alan Hindmarsh, Radu Serban,
#  *                Aaron Collier, and Slaven Peles @ LLNL
#  * -----------------------------------------------------------------
#  * SUNDIALS Copyright Start
#  * Copyright (c) 2002-2019, Lawrence Livermore National Security
#  * and Southern Methodist University.
#  * All rights reserved.
#  *
#  * See the top-level LICENSE and NOTICE files for details.
#  *
#  * SPDX-License-Identifier: BSD-3-Clause
#  * SUNDIALS Copyright End
#  * -----------------------------------------------------------------
#  * This header file exports three types: realtype, sunindextype and
#  * booleantype, as well as the constants SUNTRUE and SUNFALSE.
#  *
#  * Users should include the header file sundials_types.h in every
#  * program file and use the exported name realtype instead of
#  * float, double or long double.
#  *
#  * The constants SUNDIALS_SINGLE_PRECISION, SUNDIALS_DOUBLE_PRECISION
#  * and SUNDIALS_LONG_DOUBLE_PRECISION indicate the underlying data
#  * type of realtype.
#  *
#  * The legal types for realtype are float, double and long double.
#  *
#  * The constants SUNDIALS_INT64_T and SUNDIALS_INT32_T indicate
#  * the underlying data type of sunindextype -- the integer data type
#  * used for vector and matrix indices.
#  *
#  * Data types are set at the configuration stage.
#  *
#  * The macro RCONST gives the user a convenient way to define
#  * real-valued literal constants. To use the constant 1.0, for example,
#  * the user should write the following:
#  *
#  *   #define ONE RCONST(1.0)
#  *
#  * If realtype is defined as a double, then RCONST(1.0) expands
#  * to 1.0. If realtype is defined as a float, then RCONST(1.0)
#  * expands to 1.0F. If realtype is defined as a long double,
#  * then RCONST(1.0) expands to 1.0L. There is never a need to
#  * explicitly cast 1.0 to (realtype). The macro can be used for
#  * literal constants only. It cannot be used for expressions.
#  * -----------------------------------------------------------------*/
#
# #ifndef _SUNDIALSTYPES_H
# #define _SUNDIALSTYPES_H
#
# #ifndef _SUNDIALS_CONFIG_H
# #define _SUNDIALS_CONFIG_H
# #include <sundials/sundials_config.h>
# #endif
#
# #include <float.h>
# #include <stddef.h>
# #include <stdint.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
# /*
#  *------------------------------------------------------------------
#  * Macro _SUNDIALS_STRUCT_
#  * The _SUNDIALS_STRUCT_ macro is defined as a `struct` unless
#  * generating the SWIG interfaces - in that case it is defined as
#  * nothing. This is needed to work around a bug in SWIG which prevents
#  * it from properly parsing our generic module structures.
#  *------------------------------------------------------------------
#  */
# #ifdef SWIG
# #define _SUNDIALS_STRUCT_
# #else
# #define _SUNDIALS_STRUCT_ struct
# #endif
#
# /*
#  *------------------------------------------------------------------
#  * Type realtype
#  * Macro RCONST
#  * Constants BIG_REAL, SMALL_REAL, and UNIT_ROUNDOFF
#  *------------------------------------------------------------------
#  */
#
# #if defined(SUNDIALS_SINGLE_PRECISION)
#
# typedef float realtype;
realtype = c_double
# # define RCONST(x) x##F
# # define BIG_REAL FLT_MAX
# # define SMALL_REAL FLT_MIN
# # define UNIT_ROUNDOFF FLT_EPSILON
#
# #elif defined(SUNDIALS_DOUBLE_PRECISION)
#
# typedef double realtype;
# # define RCONST(x) x
# # define BIG_REAL DBL_MAX
# # define SMALL_REAL DBL_MIN
# # define UNIT_ROUNDOFF DBL_EPSILON
#
# #elif defined(SUNDIALS_EXTENDED_PRECISION)
#
# typedef long double realtype;
# # define RCONST(x) x##L
# # define BIG_REAL LDBL_MAX
# # define SMALL_REAL LDBL_MIN
# # define UNIT_ROUNDOFF LDBL_EPSILON
#
# #endif
#
#
# /*
#  *------------------------------------------------------------------
#  * Type : sunindextype
#  *------------------------------------------------------------------
#  * Defines integer type to be used for vector and matrix indices.
#  * User can build sundials to use 32- or 64-bit signed integers.
#  * If compiler does not support portable data types, the SUNDIALS
#  * CMake build system tries to find a type of the desired size.
#  *------------------------------------------------------------------
#  */
#
# typedef SUNDIALS_INDEX_TYPE sunindextype;
sunindextype = c_int64
#
# /*
#  *------------------------------------------------------------------
#  * Type : booleantype
#  *------------------------------------------------------------------
#  * Constants : SUNFALSE and SUNTRUE
#  *------------------------------------------------------------------
#  * ANSI C does not have a built-in boolean data type. Below is the
#  * definition for a new type called booleantype. The advantage of
#  * using the name booleantype (instead of int) is an increase in
#  * code readability. It also allows the programmer to make a
#  * distinction between int and boolean data. Variables of type
#  * booleantype are intended to have only the two values SUNFALSE and
#  * SUNTRUE which are defined below to be equal to 0 and 1,
#  * respectively.
#  *------------------------------------------------------------------
#  */
#
# #ifndef booleantype
# #define booleantype int
# #endif
booleantype = c_int
#
#
# #ifndef SUNFALSE
# #define SUNFALSE 0
# #endif
#
# #ifndef SUNTRUE
# #define SUNTRUE 1
# #endif
#
# #ifdef __cplusplus
# }
# #endif
#
# #endif  /* _SUNDIALSTYPES_H */


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'D:\a\1\s\fmpy\gui\forms\AboutDialog.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(474, 234)
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        Dialog.setPalette(palette)
        Dialog.setModal(True)
        self.formLayout = QtWidgets.QFormLayout(Dialog)
        self.formLayout.setContentsMargins(50, 50, 50, 9)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.fmpyVersionLabel = QtWidgets.QLabel(Dialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.fmpyVersionLabel.setFont(font)
        self.fmpyVersionLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.fmpyVersionLabel.setObjectName("fmpyVersionLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.fmpyVersionLabel)
        self.label_10 = QtWidgets.QLabel(Dialog)
        self.label_10.setObjectName("label_10")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_10)
        self.fmiPlatformLabel = QtWidgets.QLabel(Dialog)
        self.fmiPlatformLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.fmiPlatformLabel.setObjectName("fmiPlatformLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.fmiPlatformLabel)
        self.label_6 = QtWidgets.QLabel(Dialog)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.installationPathLabel = QtWidgets.QLabel(Dialog)
        self.installationPathLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.installationPathLabel.setObjectName("installationPathLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.installationPathLabel)
        self.label_4 = QtWidgets.QLabel(Dialog)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.pythonInterpreterLabel = QtWidgets.QLabel(Dialog)
        self.pythonInterpreterLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.pythonInterpreterLabel.setObjectName("pythonInterpreterLabel")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.pythonInterpreterLabel)
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.pythonVersionLabel = QtWidgets.QLabel(Dialog)
        self.pythonVersionLabel.setWordWrap(True)
        self.pythonVersionLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.pythonVersionLabel.setObjectName("pythonVersionLabel")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.pythonVersionLabel)
        spacerItem = QtWidgets.QSpacerItem(20, 60, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(8, QtWidgets.QFormLayout.FieldRole, spacerItem)
        self.label_8 = QtWidgets.QLabel(Dialog)
        self.label_8.setObjectName("label_8")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.FieldRole, self.label_8)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "About FMPy"))
        self.label.setText(_translate("Dialog", "FMPy Version"))
        self.fmpyVersionLabel.setText(_translate("Dialog", "2.0.0"))
        self.label_10.setText(_translate("Dialog", "FMI Platform"))
        self.fmiPlatformLabel.setText(_translate("Dialog", "win64"))
        self.label_6.setText(_translate("Dialog", "Installation Path"))
        self.installationPathLabel.setText(_translate("Dialog", "C:\\Anaconda3\\lib\\site-packages\\fmpy"))
        self.label_4.setText(_translate("Dialog", "Python Interpreter"))
        self.pythonInterpreterLabel.setText(_translate("Dialog", "C:\\Anaconda3\\python.exe"))
        self.label_2.setText(_translate("Dialog", "Python Version"))
        self.pythonVersionLabel.setText(_translate("Dialog", "3.5.2 |Anaconda custom (64-bit)| (default, Jul  5 2016, 11:41:13) [MSC v.1900 64 bit (AMD64)]"))
        self.label_8.setText(_translate("Dialog", "© 2020 Dassault Systèmes. All rights reserved."))



print("import: 'fmpy'")
import fmpy

print("import: 'fmpy.cross_check'")
import fmpy.cross_check

print("import: 'fmpy.examples'")
import fmpy.examples

print("import: 'fmpy.fmucontainer'")
import fmpy.fmucontainer

print("import: 'fmpy.ssp'")
import fmpy.ssp

print("import: 'fmpy.sundials'")
import fmpy.sundials




from ctypes import *

# /* -----------------------------------------------------------------
#  * Programmer(s): Daniel Reynolds @ SMU
#  *                David Gardner, Carol Woodward, Slaven Peles @ LLNL
#  * -----------------------------------------------------------------
#  * SUNDIALS Copyright Start
#  * Copyright (c) 2002-2019, Lawrence Livermore National Security
#  * and Southern Methodist University.
#  * All rights reserved.
#  *
#  * See the top-level LICENSE and NOTICE files for details.
#  *
#  * SPDX-License-Identifier: BSD-3-Clause
#  * SUNDIALS Copyright End
#  * -----------------------------------------------------------------
#  * This is the header file for a generic linear solver package.
#  * It defines the SUNLinearSolver structure (_generic_SUNLinearSolver)
#  * which contains the following fields:
#  *   - an implementation-dependent 'content' field which contains
#  *     any internal data required by the solver
#  *   - an 'ops' filed which contains a structure listing operations
#  *     acting on/by such solvers
#  *
#  * We consider both direct linear solvers and iterative linear solvers
#  * as available implementations of this package.  Furthermore, iterative
#  * linear solvers can either use a matrix or be matrix-free.  As a
#  * result of these different solver characteristics, some of the
#  * routines are applicable only to some types of linear solver.
#  * -----------------------------------------------------------------
#  * This header file contains:
#  *   - enumeration constants for all SUNDIALS-defined linear solver
#  *     types, as well as a generic type for user-supplied linear
#  *     solver types,
#  *   - type declarations for the _generic_SUNLinearSolver and
#  *     _generic_SUNLinearSolver_Ops structures, as well as references
#  *     to pointers to such structures (SUNLinearSolver),
#  *   - prototypes for the linear solver functions which operate
#  *     on/by SUNLinearSolver objects, and
#  *   - return codes for SUNLinearSolver objects.
#  * -----------------------------------------------------------------
#  * At a minimum, a particular implementation of a SUNLinearSolver must
#  * do the following:
#  *   - specify the 'content' field of SUNLinearSolver,
#  *   - implement the operations on/by those SUNLinearSolver objects,
#  *   - provide a constructor routine for new SUNLinearSolver objects
#  *
#  * Additionally, a SUNLinearSolver implementation may provide the
#  * following:
#  *   - "Set" routines to control solver-specific parameters/options
#  *   - "Get" routines to access solver-specific performance metrics
#  * -----------------------------------------------------------------*/
#
# #ifndef _SUNLINEARSOLVER_H
# #define _SUNLINEARSOLVER_H
#
# #include <sundials/sundials_types.h>
# #include <sundials/sundials_iterative.h>
# #include <sundials/sundials_matrix.h>
# #include <sundials/sundials_nvector.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
#
# /* -----------------------------------------------------------------
#  * Implemented SUNLinearSolver types and IDs:
#  * ----------------------------------------------------------------- */
#
# typedef enum {
#   SUNLINEARSOLVER_DIRECT,
#   SUNLINEARSOLVER_ITERATIVE,
#   SUNLINEARSOLVER_MATRIX_ITERATIVE
# } SUNLinearSolver_Type;
#
# typedef enum {
#   SUNLINEARSOLVER_BAND,
#   SUNLINEARSOLVER_DENSE,
#   SUNLINEARSOLVER_KLU,
#   SUNLINEARSOLVER_LAPACKBAND,
#   SUNLINEARSOLVER_LAPACKDENSE,
#   SUNLINEARSOLVER_PCG,
#   SUNLINEARSOLVER_SPBCGS,
#   SUNLINEARSOLVER_SPFGMR,
#   SUNLINEARSOLVER_SPGMR,
#   SUNLINEARSOLVER_SPTFQMR,
#   SUNLINEARSOLVER_SUPERLUDIST,
#   SUNLINEARSOLVER_SUPERLUMT,
#   SUNLINEARSOLVER_CUSOLVERSP_BATCHQR,
#   SUNLINEARSOLVER_CUSTOM
# } SUNLinearSolver_ID;
#
#
# /* -----------------------------------------------------------------
#  * Generic definition of SUNLinearSolver
#  * ----------------------------------------------------------------- */
#
# /* Forward reference for pointer to SUNLinearSolver_Ops object */
# typedef _SUNDIALS_STRUCT_ _generic_SUNLinearSolver_Ops *SUNLinearSolver_Ops;
#
# /* Forward reference for pointer to SUNLinearSolver object */
# typedef _SUNDIALS_STRUCT_ _generic_SUNLinearSolver *SUNLinearSolver;
SUNLinearSolver = c_void_p
#
# /* Structure containing function pointers to linear solver operations */
# struct _generic_SUNLinearSolver_Ops {
#   SUNLinearSolver_Type (*gettype)(SUNLinearSolver);
#   SUNLinearSolver_ID   (*getid)(SUNLinearSolver);
#   int                  (*setatimes)(SUNLinearSolver, void*, ATimesFn);
#   int                  (*setpreconditioner)(SUNLinearSolver, void*,
#                                             PSetupFn, PSolveFn);
#   int                  (*setscalingvectors)(SUNLinearSolver,
#                                             N_Vector, N_Vector);
#   int                  (*initialize)(SUNLinearSolver);
#   int                  (*setup)(SUNLinearSolver, SUNMatrix);
#   int                  (*solve)(SUNLinearSolver, SUNMatrix, N_Vector,
#                                 N_Vector, realtype);
#   int                  (*numiters)(SUNLinearSolver);
#   realtype             (*resnorm)(SUNLinearSolver);
#   sunindextype         (*lastflag)(SUNLinearSolver);
#   int                  (*space)(SUNLinearSolver, long int*, long int*);
#   N_Vector             (*resid)(SUNLinearSolver);
#   int                  (*free)(SUNLinearSolver);
# };
#
# /* A linear solver is a structure with an implementation-dependent
#    'content' field, and a pointer to a structure of linear solver
#    operations corresponding to that implementation. */
# struct _generic_SUNLinearSolver {
#   void *content;
#   SUNLinearSolver_Ops ops;
# };
#
#
# /* -----------------------------------------------------------------
#  * Functions exported by SUNLinearSolver module
#  * ----------------------------------------------------------------- */
#
# SUNDIALS_EXPORT SUNLinearSolver SUNLinSolNewEmpty();
#
# SUNDIALS_EXPORT void SUNLinSolFreeEmpty(SUNLinearSolver S);
#
# SUNDIALS_EXPORT SUNLinearSolver_Type SUNLinSolGetType(SUNLinearSolver S);
#
# SUNDIALS_EXPORT SUNLinearSolver_ID SUNLinSolGetID(SUNLinearSolver S);
#
# SUNDIALS_EXPORT int SUNLinSolSetATimes(SUNLinearSolver S, void* A_data,
#                                        ATimesFn ATimes);
#
# SUNDIALS_EXPORT int SUNLinSolSetPreconditioner(SUNLinearSolver S, void* P_data,
#                                                PSetupFn Pset, PSolveFn Psol);
#
# SUNDIALS_EXPORT int SUNLinSolSetScalingVectors(SUNLinearSolver S, N_Vector s1,
#                                                N_Vector s2);
#
# SUNDIALS_EXPORT int SUNLinSolInitialize(SUNLinearSolver S);
#
# SUNDIALS_EXPORT int SUNLinSolSetup(SUNLinearSolver S, SUNMatrix A);
#
# SUNDIALS_EXPORT int SUNLinSolSolve(SUNLinearSolver S, SUNMatrix A, N_Vector x,
#                                    N_Vector b, realtype tol);
#
# SUNDIALS_EXPORT int SUNLinSolNumIters(SUNLinearSolver S);
#
# SUNDIALS_EXPORT realtype SUNLinSolResNorm(SUNLinearSolver S);
#
# SUNDIALS_EXPORT N_Vector SUNLinSolResid(SUNLinearSolver S);
#
# SUNDIALS_EXPORT sunindextype SUNLinSolLastFlag(SUNLinearSolver S);
#
# SUNDIALS_EXPORT int SUNLinSolSpace(SUNLinearSolver S, long int *lenrwLS,
#                                    long int *leniwLS);
#
# SUNDIALS_EXPORT int SUNLinSolFree(SUNLinearSolver S);
#
#
# /* -----------------------------------------------------------------
#  * SUNLinearSolver return values
#  * ----------------------------------------------------------------- */
#
# #define SUNLS_SUCCESS               0   /* successful/converged          */
#
# #define SUNLS_MEM_NULL           -801   /* mem argument is NULL          */
# #define SUNLS_ILL_INPUT          -802   /* illegal function input        */
# #define SUNLS_MEM_FAIL           -803   /* failed memory access          */
# #define SUNLS_ATIMES_FAIL_UNREC  -804   /* atimes unrecoverable failure  */
# #define SUNLS_PSET_FAIL_UNREC    -805   /* pset unrecoverable failure    */
# #define SUNLS_PSOLVE_FAIL_UNREC  -806   /* psolve unrecoverable failure  */
# #define SUNLS_PACKAGE_FAIL_UNREC -807   /* external package unrec. fail  */
# #define SUNLS_GS_FAIL            -808   /* Gram-Schmidt failure          */
# #define SUNLS_QRSOL_FAIL         -809   /* QRsol found singular R        */
# #define SUNLS_VECTOROP_ERR       -810   /* vector operation error        */
#
# #define SUNLS_RES_REDUCED         801   /* nonconv. solve, resid reduced */
# #define SUNLS_CONV_FAIL           802   /* nonconvergent solve           */
# #define SUNLS_ATIMES_FAIL_REC     803   /* atimes failed recoverably     */
# #define SUNLS_PSET_FAIL_REC       804   /* pset failed recoverably       */
# #define SUNLS_PSOLVE_FAIL_REC     805   /* psolve failed recoverably     */
# #define SUNLS_PACKAGE_FAIL_REC    806   /* external package recov. fail  */
# #define SUNLS_QRFACT_FAIL         807   /* QRfact found singular matrix  */
# #define SUNLS_LUFACT_FAIL         808   /* LUfact found singular matrix  */
#
# #ifdef __cplusplus
# }
# #endif
# #endif



# noinspection PyPep8

import shutil
from .fmi1 import *
from .fmi1 import _FMU1
from .fmi2 import *
from .fmi2 import _FMU2
from . import fmi3
from . import extract
from .util import auto_interval, _is_string
import numpy as np
from time import time as current_time

# absolute tolerance for equality when comparing two floats
eps = 1e-13


class SimulationResult(np.ndarray):

    def __new__(subtype, shape, dtype=float, buffer=None, offset=0, strides=None, order=None, modelDescription=None):
        obj = super(SimulationResult, subtype).__new__(subtype, shape, dtype, buffer, offset, strides, order)
        obj.modelDescription = modelDescription
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.modelDescription = getattr(obj, 'modelDescription', None)


def _get_output_variables(model_description, max_variables=5):
    """ Create a list of default output variables """

    output_variables = []

    # output variables
    for variable in model_description.modelVariables:
        if variable.causality == 'output':
            output_variables.append(variable)

    if len(output_variables) > 0:
        return output_variables

    # continuous states
    if model_description.derivatives is not None:
        output_variables = [derivative.variable.derivative for derivative in model_description.derivatives]

    if len(output_variables) > 0:
        return output_variables[:max_variables]

    # local variables
    for variable in model_description.modelVariables:
        if variable.variability == 'local':
            output_variables.append(variable)

    if len(output_variables) > 0:
        return output_variables[:max_variables]

    # any variable
    return model_description.modelVariables[:max_variables]


class Recorder(object):
    """ Helper class to record the variables during the simulation """

    def __init__(self, fmu, modelDescription, variableNames=None, interval=None):
        """
        Parameters:
            fmu               the FMU instance
            modelDescription  the model description instance
            variableNames     list of variable names to record
            interval          minimum distance to the previous sample
        """

        self.fmu = fmu
        self.interval = interval

        self.cols = [('time', np.float64, None)]  # name, dtype, shape
        self.rows = []
        self.info = {}  # type -> (names, vrs, shapes, n_values, getter)
        self.types = []

        self.constants = {}
        self.modelDescription = modelDescription

        if variableNames is None:
            variableNames = [variable.name for variable in _get_output_variables(modelDescription)]

        # collect the variables to record
        for sv in modelDescription.modelVariables:

            if sv.name == 'time':
                continue  # "time" is reserved for the simulation time

            # collect the variables to record
            if sv.name in variableNames:
                type = sv.type
                if type == 'Enumeration':
                    type = 'Integer' if modelDescription.fmiVersion in {'1.0', '2.0'} else 'Int32'
                names, vrs, shapes, n_values, getter = self.info.get(type, ([], [], [], 0, getattr(self.fmu, 'get' + type)))
                names.append(sv.name)
                vrs.append(sv.valueReference)
                shapes.append(sv.shape)
                n_values += np.prod(sv.shape) if sv.shape else 1
                self.info[type] = names, vrs, shapes, n_values, getter

        # create the columns for the NumPy array
        if modelDescription.fmiVersion in ['1.0', '2.0']:
            types = [('Real', np.float64), ('Integer', np.int32), ('Boolean', np.bool_)]
        else:
            types = [('Float64', np.float64), ('Int32', np.int32), ('UInt64', np.uint64), ('Boolean', np.bool_)]

        # collect the columns
        for t, dt in types:
            if t in self.info:
                self.types.append(t)
                names, _, shapes, _, _ = self.info[t]
                self.cols += zip(names, [dt] * len(names), shapes)

        # strip the shape for scalars
        self.cols = [(n, t) if not s else (n, t, s) for n, t, s in self.cols]

    @staticmethod
    def _append_reshaped(row, values, shapes):
        i = 0
        for d in shapes:
            if d:
                s = np.prod(d)
                value = np.array(values[i:i + s]).reshape(d)
                i += s
            else:
                value = values[i]
                i += 1
            row.append(value)

    def sample(self, time, force=False):
        """ Record the variables """

        if not force and self.interval is not None and len(self.rows) > 0:
            last = self.rows[-1][0]
            if time - last + eps < self.interval:
                return

        row = [time]

        for t in self.types:
            names, vrs, shapes, nValues, getter = self.info[t]
            if self.modelDescription.fmiVersion in ['1.0', '2.0']:
                values = getter(vr=vrs)
            else:
                values = getter(vr=vrs, nValues=nValues)
            self._append_reshaped(row, values, shapes)

        self.rows.append(tuple(row))

    def result(self):
        """ Return a structured NumPy array with the recorded results """

        arr = np.array(self.rows, dtype=np.dtype(self.cols))

        info_arr = arr.view(SimulationResult)

        info_arr.modelDescription = self.modelDescription

        return info_arr

    @property
    def lastSampleTime(self):
        """ Return the last sample time """

        if len(self.rows) > 0:
            return self.rows[-1][0]
        raise Exception("No samples available")


class Input(object):
    """ Helper class that sets the input to the FMU """

    def __init__(self, fmu, modelDescription, signals):
        """
        Parameters:
            fmu               the FMU instance
            modelDescription  the model description instance
            signals           a structured numpy array that contains the input

        Example:

        Create a Real signal 'step' and a Boolean signal 'switch' with a discrete step at t=0.5

        >>> import numpy as np
        >>> dtype = [('time', np.double), ('step', np.double), ('switch', np.bool_)]
        >>> signals = np.array([(0.0, 0.0, False), (0.5, 0.0, False), (0.5, 0.1, True), (1.0, 1.0, True)], dtype=dtype)
        """

        self.fmu = fmu

        if signals is None:
            self.t = None
            return

        # get the time grid
        self.t = signals[signals.dtype.names[0]]

        # find events
        self.t_events = Input.findEvents(signals, modelDescription)

        is_fmi1 = isinstance(fmu, _FMU1)
        is_fmi2 = isinstance(fmu, _FMU2)

        self.set_input_derivatives = is_fmi2 and modelDescription.coSimulation and modelDescription.coSimulation.canInterpolateInputs

        setters = dict()

        # get the setters
        if is_fmi1:
            setters['Real']    = (fmu.fmi1SetReal,    fmi1Real)
            setters['Integer'] = (fmu.fmi1SetInteger, fmi1Integer)
            setters['Boolean'] = (fmu.fmi1SetBoolean, c_int8)
        elif is_fmi2:
            setters['Real']    = (fmu.fmi2SetReal,    fmi2Real)
            setters['Integer'] = (fmu.fmi2SetInteger, fmi2Integer)
            setters['Boolean'] = (fmu.fmi2SetBoolean, fmi2Boolean)
        else:
            setters['Float64'] = (fmu.fmi3SetFloat64, fmi3.fmi3Float64)
            setters['Int32']   = (fmu.fmi3SetInt32,   fmi3.fmi3Int32)
            setters['Boolean'] = (fmu.fmi3SetBoolean, fmi3.fmi3Boolean)

        from collections import defaultdict

        continuous_inputs = defaultdict(list)
        discrete_inputs = defaultdict(list)

        self.continuous = []
        self.discrete = []

        for variable in modelDescription.modelVariables:

            if variable.causality != 'input' and variable.variability != 'tunable':
                continue

            if variable.name not in signals.dtype.names:
                if variable.causality == 'input':
                    print('Warning: missing input for variable "%s"' % variable.name)
                continue

            if variable.type in {'Float32', 'Float64', 'Real'} and variable.variability not in ['discrete', 'tunable']:
                continuous_inputs[variable.type].append((variable.valueReference, variable.name))
            else:
                # use the same table for Integer and Enumeration
                type_ = 'Integer' if variable.type == 'Enumeration' else variable.type
                discrete_inputs[type_].append((variable.valueReference, variable.name))

        for type_, vrs_and_names in continuous_inputs.items():
            vrs, names = zip(*vrs_and_names)
            setter, value_type = setters[type_]
            self.continuous.append((
                (c_uint32 * len(vrs))(*vrs),
                (value_type * len(vrs))(),
                (c_int * len(vrs))(*([1] * len(vrs))),
                (value_type * len(vrs))(),
                np.asarray(np.stack(list(map(lambda n: signals[n], names))), dtype=value_type),
                setter
            ))

        for type_, vrs_and_names in discrete_inputs.items():
            vrs, names = zip(*vrs_and_names)
            setter, value_type = setters[type_]
            self.discrete.append((
                (c_uint32 * len(vrs))(*vrs),
                (value_type * len(vrs))(),
                np.asarray(np.stack(list(map(lambda n: signals[n], names))), dtype=value_type),
                setter
            ))

    def apply(self, time, continuous=True, discrete=True, after_event=False):
        """ Apply the input

        Parameters:
            continuous   apply continuous inputs
            discrete     apply discrete inputs
            after_event  apply right hand side inputs at discontinuities
        """

        if self.t is None:
            return

        is_fmi3 = isinstance(self.fmu, fmi3._FMU3)

        # continuous
        if continuous:
            for vrs, values, order, derivatives, table, setter in self.continuous:
                values[:], derivatives[:] = self.interpolate(time=time, t=self.t, table=table, discrete=False, after_event=after_event)
                if is_fmi3:
                    setter(self.fmu.component, vrs, len(vrs), values, len(values))
                else:
                    setter(self.fmu.component, vrs, len(vrs), values)

                if self.set_input_derivatives:
                    self.fmu.fmi2SetRealInputDerivatives(self.fmu.component, vrs, len(vrs), order, derivatives)

        # discrete
        if discrete:
            for vrs, values, table, setter in self.discrete:
                values[:], der_values = self.interpolate(time=time, t=self.t, table=table, discrete=True, after_event=after_event)

                if values._type_ == c_int8:
                    # special treatment for fmi1Boolean
                    setter(self.fmu.component, vrs, len(vrs), cast(values, POINTER(c_char)))
                else:
                    if is_fmi3:
                        setter(self.fmu.component, vrs, len(vrs), values, len(values))
                    else:
                        setter(self.fmu.component, vrs, len(vrs), values)

    def nextEvent(self, time):
        """ Get the next input event """

        if self.t is None:
            return float('Inf')

        # find the next event
        i = np.argmax(self.t_events > time)
        return self.t_events[i]

    @staticmethod
    def findEvents(signals, model_description):
        """ Find time events """

        t_event = {float('Inf')}

        if signals.size < 2:
            return np.array(list(t_event))  # only one sample

        t = signals[signals.dtype.names[0]]

        # continuous
        i_event = np.where(np.diff(t) == 0)
        t_event.update(t[i_event])

        # discrete
        for variable in model_description.modelVariables:
            if variable.name in signals.dtype.names and variable.variability in ['discrete', 'tunable']:
                y = signals[variable.name]
                i_event = np.flatnonzero(np.diff(y))
                t_event.update(t[i_event + 1])

        return np.array(sorted(t_event))

    @staticmethod
    def interpolate(time, t, table, discrete=False, after_event=False):

        if t.size < 2:
            return table, np.zeros_like(table)  # only one sample

        # find the left insert index
        i0 = np.searchsorted(t, time)

        if i0 == 0:
            values = table[:, 0]  # hold first value
            return values, np.zeros_like(values)

        if i0 == len(t):
            values = table[:, -1]  # hold last value
            return values, np.zeros_like(values)

        # check for event
        if time == t[i0] and i0 < len(t) - 1 and t[i0] == t[i0 + 1]:

            if after_event:
                # take the value after the event
                while i0 < len(t) - 1 and t[i0] == t[i0 + 1]:
                    i0 += 1
                if i0 < len(t) - 1:
                    v0 = table[:, i0]
                    v1 = table[:, i0 + 1]
                    t0 = t[i0]
                    t1 = t[i0 + 1]
                    der_v = (v1 - v0) / (t1 - t0)
                else:
                    der_v = np.zeros((table.shape[0],))
            else:
                v0 = table[:, i0 - 1]
                v1 = table[:, i0]
                t0 = t[i0 - 1]
                t1 = t[i0]
                der_v = (v1 - v0) / (t1 - t0)

            values = table[:, i0]
            return values, der_v

        i0 -= 1  # interpolate
        i1 = i0 + 1

        if discrete:
            values = table[:, i1 if after_event else i0]
            return values, np.zeros_like(values)

        t0 = t[i0]
        t1 = t[i1]

        w0 = (t1 - time) / (t1 - t0)
        w1 = 1 - w0

        v0 = table[:, i0]
        v1 = table[:, i1]

        # interpolate the input value
        v = w0 * v0 + w1 * v1
        der_v = (v1 - v0) / (t1 - t0)

        return v, der_v


def apply_start_values(fmu, model_description, start_values, apply_default_start_values=False):
    """ Set start values to an FMU instance

    Parameters:
        fmu                     the FMU instance
        model_description       the ModelDescription instance
        start_values            dictionary of variable_name -> start_value pairs
        apply_default_values    apply the start values from the model description
    """

    start_values = start_values.copy()

    unit_definitions = dict((u.name, dict((d.name, d) for d in u.displayUnits)) for u in model_description.unitDefinitions)

    for variable in model_description.modelVariables:

        if variable.name in start_values:
            value = start_values.pop(variable.name)
        elif apply_default_start_values and variable.start is not None:
            value = variable.start
        else:
            continue

        if type(value) is tuple:
            if len(value) != 2:
                raise Exception('The start value for variable %s must be a scalar value or a tuple (<value>, {<unit>|<display_unit>}) but was "%s".' % (variable.name, value))
            value, unit = value
        else:
            unit = None

        if unit is None or unit == variable.unit:
            pass
        elif variable.declaredType is not None and unit == variable.declaredType.unit:
            pass
        else:
            if variable.unit is not None:
                base_unit = variable.unit
            elif variable.declaredType is not None:
                base_unit = variable.declaredType.unit
            else:
                raise Exception('Variable %s has no unit but the unit "%s" was specified for its start value.' % (variable.name, unit))

            if unit not in unit_definitions[base_unit]:
                raise Exception('The unit "%s" of the start value for variable %s is not defined.' % (unit, variable.name))

            display_unit = unit_definitions[base_unit][unit]
            value = (value - display_unit.offset) / display_unit.factor

        vr = variable.valueReference

        # get the setter function
        if variable.type == 'Enumeration':
            if model_description.fmiVersion in {'1.0', '2.0'}:
                setter = getattr(fmu, 'setInteger')
            else:
                setter = getattr(fmu, 'setInt32')
        else:
            setter = getattr(fmu, 'set' + variable.type)

        # convert Boolean values
        if variable.type == 'Boolean':
            if isinstance(value, str):
                if value.lower() not in ['true', 'false']:
                    raise Exception('The start value "%s" for variable "%s" could not be converted to Boolean' %
                                    (value, variable.name))
                else:
                    value = value.lower() == 'true'

        # convert the type
        if variable.shape:
            if _is_string(value):
                value = value.split()
            value = list(map(lambda e: variable._python_type(e), value))
            if len(value) != np.prod(variable.shape):
                raise ArgumentError('The start value for variable "%s" must have %d elements but has %d.' %
                                    (variable.name, np.prod(variable.shape), len(value)))
        else:
            value = [variable._python_type(value)]

        setter([vr], value)

    if len(start_values) > 0:
        raise Exception("The start values for the following variables could not be set because they don't exist: " +
                        ', '.join(start_values.keys()))


class ForwardEuler(object):

    def __init__(self, nx, nz, get_x, set_x, get_dx, get_z, input):

        self.get_x = get_x
        self.set_x = set_x
        self.get_dx = get_dx
        self.get_z = get_z

        self.x = np.zeros(nx)
        self.dx = np.zeros(nx)
        self.z = np.zeros(nz)
        self.prez = np.zeros(nz)

        self._px = self.x.ctypes.data_as(POINTER(c_double))
        self._pdx = self.dx.ctypes.data_as(POINTER(c_double))
        self._pz = self.z.ctypes.data_as(POINTER(c_double))
        self._pprez = self.z.ctypes.data_as(POINTER(c_double))

        # initialize the event indicators
        self.get_z(self._pz, self.z.size)

    def step(self, t, tNext):

        # get the current states and derivatives
        self.get_x(self._px, self.x.size)
        self.get_dx(self._pdx, self.dx.size)

        # perform one step
        dt = tNext - t
        self.x += dt * self.dx

        # set the continuous states
        self.set_x(self._px, self.x.size)

        # check for state event
        self.prez[:] = self.z
        self.get_z(self._pz, self.z.size)

        roots = np.zeros_like(self.z, dtype=np.int32)

        # find zero crossings
        for i, (prez, z) in enumerate(zip(self.prez, self.z)):
            if prez < 0 and z >= 0:
                roots[i] = -1
            elif prez > 0 and z <= 0:
                roots[i] = 1

        return np.any(roots != 0), roots, tNext

    def reset(self, time):
        pass  # nothing to do


def simulate_fmu(filename,
                 validate=True,
                 start_time=None,
                 stop_time=None,
                 solver='CVode',
                 step_size=None,
                 relative_tolerance=None,
                 output_interval=None,
                 record_events=True,
                 fmi_type=None,
                 start_values={},
                 apply_default_start_values=False,
                 input=None,
                 output=None,
                 timeout=None,
                 debug_logging=False,
                 visible=False,
                 logger=None,
                 fmi_call_logger=None,
                 step_finished=None,
                 model_description=None,
                 fmu_instance=None):
    """ Simulate an FMU

    Parameters:
        filename            filename of the FMU or directory with extracted FMU
        validate            validate the FMU
        start_time          simulation start time (None: use default experiment or 0 if not defined)
        stop_time           simulation stop time (None: use default experiment or start_time + 1 if not defined)
        solver              solver to use for model exchange ('Euler' or 'CVode')
        step_size           step size for the 'Euler' solver
        relative_tolerance  relative tolerance for the 'CVode' solver and FMI 2.0 co-simulation FMUs
        output_interval     interval for sampling the output
        record_events       record outputs at events (model exchange only)
        fmi_type            FMI type for the simulation (None: determine from FMU)
        start_values        dictionary of variable name -> value pairs
        apply_default_start_values  apply the start values from the model description
        input               a structured numpy array that contains the input (see :class:`Input`)
        output              list of variables to record (None: record outputs)
        timeout             timeout for the simulation
        debug_logging       enable the FMU's debug logging
        visible             interactive mode (True) or batch mode (False)
        fmi_call_logger     callback function to log FMI calls
        logger              callback function passed to the FMU (experimental)
        step_finished       callback to interact with the simulation (experimental)
        model_description   the previously loaded model description (experimental)
        fmu_instance        the previously instantiated FMU (experimental)

    Returns:
        result              a structured numpy array that contains the result
    """

    from fmpy import supported_platforms, platform_tuple
    from fmpy.model_description import read_model_description

    platforms = supported_platforms(filename)

    # use 32-bit DLL remoting
    use_remoting = platform == 'win64' and 'win64' not in platforms and 'win32' in platforms

    if fmu_instance is None and platform not in platforms and not use_remoting:
        raise Exception("The current platform (%s) is not supported by the FMU." % platform)

    if model_description is None:
        model_description = read_model_description(filename, validate=validate)
    else:
        model_description = model_description

    if fmi_type is None:
        if fmu_instance is not None:
            # determine FMI type from the FMU instance
            fmi_type = 'CoSimulation' if type(fmu_instance) in [FMU1Slave, FMU2Slave, fmi3.FMU3Slave] else 'ModelExchange'
        else:
            # determine the FMI type automatically
            fmi_type = 'CoSimulation' if model_description.coSimulation is not None else 'ModelExchange'

    if fmi_type not in ['ModelExchange', 'CoSimulation']:
        raise Exception('fmi_type must be one of "ModelExchange" or "CoSimulation"')

    experiment = model_description.defaultExperiment

    if start_time is None:
        if experiment is not None and experiment.startTime is not None:
            start_time = experiment.startTime
        else:
            start_time = 0.0

    start_time = float(start_time)

    if stop_time is None:
        if experiment is not None and experiment.stopTime is not None:
            stop_time = experiment.stopTime
        else:
            stop_time = start_time + 1.0

    stop_time = float(stop_time)

    if relative_tolerance is None and experiment is not None:
        relative_tolerance = experiment.tolerance

    if step_size is None:
        total_time = stop_time - start_time
        step_size = 10 ** (np.round(np.log10(total_time)) - 3)

    if output_interval is None and fmi_type == 'CoSimulation' and experiment is not None and experiment.stepSize is not None:
        output_interval = experiment.stepSize
        while (stop_time - start_time) / output_interval > 1000:
            output_interval *= 2

    if os.path.isfile(os.path.join(filename, 'modelDescription.xml')):
        unzipdir = filename
        tempdir = None
    else:
        required_paths = ['resources', 'binaries/' + platform, 'binaries/' + platform_tuple]
        if use_remoting:
            required_paths.append('binaries/win32')
        tempdir = extract(filename, include=lambda n: n.startswith(tuple(required_paths)))
        unzipdir = tempdir

    if use_remoting:
        # start 32-bit server
        from subprocess import Popen
        server_path = os.path.dirname(__file__)
        server_path = os.path.join(server_path, 'remoting', 'server.exe')
        if fmi_type == 'ModelExchange':
            model_identifier = model_description.modelExchange.modelIdentifier
        else:
            model_identifier = model_description.coSimulation.modelIdentifier
        dll_path = os.path.join(unzipdir, 'binaries', 'win32', model_identifier + '.dll')
        server = Popen([server_path, dll_path])
    else:
        server = None

    if fmu_instance is None:
        fmu = instantiate_fmu(unzipdir, model_description, fmi_type, visible, debug_logging, logger, fmi_call_logger, use_remoting)
    else:
        fmu = fmu_instance

    # simulate_fmu the FMU
    if fmi_type == 'ModelExchange':
        result = simulateME(model_description, fmu, start_time, stop_time, solver, step_size, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, record_events, timeout, step_finished)
    elif fmi_type == 'CoSimulation':
        result = simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input, output, output_interval, timeout, step_finished)

    if fmu_instance is None:
        fmu.freeInstance()

    if server is not None:
        server.kill()

    # clean up
    if tempdir is not None:
        shutil.rmtree(tempdir, ignore_errors=True)

    return result


def instantiate_fmu(unzipdir, model_description, fmi_type=None, visible=False, debug_logging=False, logger=None, fmi_call_logger=None, use_remoting=False):
    """
    Create an instance of fmpy.fmi1._FMU (see simulate_fmu() for documentation of the parameters).
    """

    # common constructor arguments
    fmu_args = {
        'guid': model_description.guid,
        'unzipDirectory': unzipdir,
        'instanceName': None,
        'fmiCallLogger': fmi_call_logger
    }

    if use_remoting:
        fmu_args['libraryPath'] = os.path.join(os.path.dirname(__file__), 'remoting', 'client.dll')

    if logger is None:
        logger = printLogMessage

    is_fmi1 = model_description.fmiVersion == '1.0'
    is_fmi2 = model_description.fmiVersion == '2.0'

    if is_fmi1:
        callbacks = fmi1CallbackFunctions()
        callbacks.logger         = fmi1CallbackLoggerTYPE(logger)
        callbacks.allocateMemory = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
        callbacks.freeMemory     = fmi1CallbackFreeMemoryTYPE(freeMemory)
        callbacks.stepFinished   = None
    elif is_fmi2:
        callbacks = fmi2CallbackFunctions()
        callbacks.logger         = fmi2CallbackLoggerTYPE(logger)
        callbacks.allocateMemory = fmi2CallbackAllocateMemoryTYPE(allocateMemory)
        callbacks.freeMemory     = fmi2CallbackFreeMemoryTYPE(freeMemory)
    else:
        callbacks = None

    if model_description.fmiVersion in ['1.0', '2.0']:
        # add native proxy function that processes variadic arguments
        try:
            from .logging import addLoggerProxy
            addLoggerProxy(byref(callbacks))
        except Exception as e:
            print("Failed to add logger proxy function. %s" % e)

    if fmi_type in [None, 'CoSimulation'] and model_description.coSimulation is not None:

        fmu_args['modelIdentifier'] = model_description.coSimulation.modelIdentifier

        if is_fmi1:
            fmu = FMU1Slave(**fmu_args)
            fmu.instantiate(functions=callbacks, loggingOn=debug_logging)
        elif is_fmi2:
            fmu = FMU2Slave(**fmu_args)
            fmu.instantiate(visible=visible, callbacks=callbacks, loggingOn=debug_logging)
        else:
            fmu = fmi3.FMU3Slave(**fmu_args)
            fmu.instantiate(visible=visible, loggingOn=debug_logging)

    elif fmi_type in [None, 'ModelExchange'] and model_description.modelExchange is not None:

        fmu_args['modelIdentifier'] = model_description.modelExchange.modelIdentifier

        if is_fmi1:
            fmu = FMU1Model(**fmu_args)
            fmu.instantiate(functions=callbacks, loggingOn=debug_logging)
        elif is_fmi2:
            fmu = FMU2Model(**fmu_args)
            fmu.instantiate(visible=visible, callbacks=callbacks, loggingOn=debug_logging)
        else:
            fmu = fmi3.FMU3Model(**fmu_args)
            fmu.instantiate(visible=visible, loggingOn=debug_logging)

    elif fmi_type in [None, 'ScheduledExecution'] and model_description.scheduledExecution is not None:

        fmu_args['modelIdentifier'] = model_description.scheduledExecution.modelIdentifier
        fmu = fmi3.FMU3ScheduledExecution(**fmu_args)
        fmu.instantiate(visible=visible, loggingOn=debug_logging)

    else:

        raise Exception('FMI type "%s" is not supported by the FMU' % fmi_type)

    return fmu


def simulateME(model_description, fmu, start_time, stop_time, solver_name, step_size, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, record_events, timeout, step_finished):

    if relative_tolerance is None:
        relative_tolerance = 1e-5

    if output_interval is None:
        if step_size is None:
            output_interval = auto_interval(stop_time - start_time)
        else:
            output_interval = step_size
            while (stop_time - start_time) / output_interval > 1000:
                output_interval *= 2

    if step_size is None:
        step_size = output_interval
        max_step = (stop_time - start_time) / 1000
        while step_size > max_step:
            step_size /= 2

    sim_start = current_time()

    time = start_time

    is_fmi1 = model_description.fmiVersion == '1.0'
    is_fmi2 = model_description.fmiVersion == '2.0'
    is_fmi3 = model_description.fmiVersion.startswith('3.0')

    if is_fmi1:
        fmu.setTime(time)
    elif is_fmi2:
        fmu.setupExperiment(startTime=start_time)

    input = Input(fmu, model_description, input_signals)

    apply_start_values(fmu, model_description, start_values, apply_default_start_values)

    # initialize
    if is_fmi1:

        input.apply(time)

        (iterationConverged,
         stateValueReferencesChanged,
         stateValuesChanged,
         terminateSimulation,
         nextEventTimeDefined,
         nextEventTime) = fmu.initialize()

        if terminateSimulation:
            raise Exception('Model requested termination during initial event update.')

    elif is_fmi2:

        fmu.enterInitializationMode()
        input.apply(time)
        fmu.exitInitializationMode()

        newDiscreteStatesNeeded = True
        terminateSimulation = False

        while newDiscreteStatesNeeded and not terminateSimulation:
            # update discrete states
            (newDiscreteStatesNeeded,
             terminateSimulation,
             nominalsOfContinuousStatesChanged,
             valuesOfContinuousStatesChanged,
             nextEventTimeDefined,
             nextEventTime) = fmu.newDiscreteStates()

        if terminateSimulation:
            raise Exception('Model requested termination during initial event update.')

        fmu.enterContinuousTimeMode()

    elif is_fmi3:

        fmu.enterInitializationMode(startTime=start_time)
        input.apply(time)
        fmu.exitInitializationMode()

        newDiscreteStatesNeeded = True
        terminateSimulation = False

        while newDiscreteStatesNeeded and not terminateSimulation:
            # update discrete states
            (newDiscreteStatesNeeded,
             terminateSimulation,
             nominalsOfContinuousStatesChanged,
             valuesOfContinuousStatesChanged,
             nextEventTimeDefined,
             nextEventTime) = fmu.newDiscreteStates()

        fmu.enterContinuousTimeMode()

    # common solver constructor arguments
    solver_args = {
        'nx': model_description.numberOfContinuousStates,
        'nz': model_description.numberOfEventIndicators,
        'get_x': fmu.getContinuousStates,
        'set_x': fmu.setContinuousStates,
        'get_dx': fmu.getDerivatives,
        'get_z': fmu.getEventIndicators,
        'input': input
    }

    # select the solver
    if solver_name == 'Euler':
        solver = ForwardEuler(**solver_args)
        fixed_step = True
    elif solver_name is None or solver_name == 'CVode':
        from .sundials import CVodeSolver
        solver = CVodeSolver(set_time=fmu.setTime,
                             startTime=start_time,
                             maxStep=(stop_time - start_time) / 50.,
                             relativeTolerance=relative_tolerance,
                             **solver_args)
        step_size = output_interval
        fixed_step = False
    else:
        raise Exception("Unknown solver: %s. Must be one of 'Euler' or 'CVode'." % solver_name)

    # check step size
    if fixed_step and not np.isclose(round(output_interval / step_size) * step_size, output_interval):
        raise Exception("output_interval must be a multiple of step_size for fixed step solvers")

    recorder = Recorder(fmu=fmu,
                        modelDescription=model_description,
                        variableNames=output,
                        interval=output_interval)

    # record the values for time == start_time
    recorder.sample(time)

    t_next = start_time

    # simulation loop
    while time < stop_time:

        if timeout is not None and (current_time() - sim_start) > timeout:
            break

        if fixed_step:
            if time + step_size < stop_time + eps:
                t_next = time + step_size
            else:
                break
        else:
            if time + eps >= t_next:  # t_next has been reached
                # integrate to the next grid point
                t_next = np.floor(time / output_interval) * output_interval + output_interval
                if t_next < time + eps:
                    t_next += output_interval

        # get the next input event
        t_input_event = input.nextEvent(time)

        # check for input event
        input_event = t_input_event <= t_next

        if input_event:
            t_next = t_input_event

        time_event = nextEventTimeDefined and nextEventTime <= t_next

        if time_event and not fixed_step:
            t_next = nextEventTime

        if t_next - time > eps:
            # do one step
            state_event, roots_found, time = solver.step(time, t_next)
        else:
            # skip
            time = t_next

        # set the time
        fmu.setTime(time)

        # apply continuous inputs
        input.apply(time, discrete=False)

        # check for step event, e.g.dynamic state selection
        if is_fmi1:
            step_event = fmu.completedIntegratorStep()
        else:
            step_event, _ = fmu.completedIntegratorStep()
            step_event = step_event != fmi2False

        # handle events
        if input_event or time_event or state_event or step_event:

            if record_events:
                # record the values before the event
                recorder.sample(time, force=True)

            if is_fmi1:

                if input_event:
                    input.apply(time=time, after_event=True)

                iterationConverged = False

                # update discrete states
                while not iterationConverged and not terminateSimulation:
                    (iterationConverged,
                     stateValueReferencesChanged,
                     stateValuesChanged,
                     terminateSimulation,
                     nextEventTimeDefined,
                     nextEventTime) = fmu.eventUpdate()

                if terminateSimulation:
                    break

            elif is_fmi2:

                fmu.enterEventMode()

                if input_event:
                    input.apply(time=time, after_event=True)

                newDiscreteStatesNeeded = True

                # update discrete states
                while newDiscreteStatesNeeded and not terminateSimulation:
                    (newDiscreteStatesNeeded,
                     terminateSimulation,
                     nominalsOfContinuousStatesChanged,
                     valuesOfContinuousStatesChanged,
                     nextEventTimeDefined,
                     nextEventTime) = fmu.newDiscreteStates()

                if terminateSimulation:
                    break

                fmu.enterContinuousTimeMode()

            else:

                fmu.enterEventMode(stepEvent=step_event, rootsFound=roots_found, timeEvent=time_event)

                if input_event:
                    input.apply(time=time, after_event=True)

                newDiscreteStatesNeeded = True

                # update discrete states
                while newDiscreteStatesNeeded and not terminateSimulation:
                    (newDiscreteStatesNeeded,
                     terminateSimulation,
                     nominalsOfContinuousStatesChanged,
                     valuesOfContinuousStatesChanged,
                     nextEventTimeDefined,
                     nextEventTime) = fmu.newDiscreteStates()

                if terminateSimulation:
                    break

                fmu.enterContinuousTimeMode()

            solver.reset(time)

            if record_events:
                # record values after the event
                recorder.sample(time, force=True)

        if abs(time - round(time / output_interval) * output_interval) < eps and time > recorder.lastSampleTime + eps:
            # record values for this step
            recorder.sample(time, force=True)

        if step_finished is not None and not step_finished(time, recorder):
            break

    fmu.terminate()

    del solver

    return recorder.result()


def simulateCS(model_description, fmu, start_time, stop_time, relative_tolerance, start_values, apply_default_start_values, input_signals, output, output_interval, timeout, step_finished):

    if output_interval is None:
        output_interval = auto_interval(stop_time - start_time)

    sim_start = current_time()

    is_fmi1 = model_description.fmiVersion == '1.0'
    is_fmi2 = model_description.fmiVersion == '2.0'

    if is_fmi2:
        fmu.setupExperiment(tolerance=relative_tolerance, startTime=start_time)

    input = Input(fmu=fmu, modelDescription=model_description, signals=input_signals)

    time = start_time

    apply_start_values(fmu, model_description, start_values, apply_default_start_values)

    # initialize the model
    if is_fmi1:
        input.apply(time)
        fmu.initialize(tStart=time, stopTime=stop_time)
    elif is_fmi2:
        fmu.enterInitializationMode()
        input.apply(time)
        fmu.exitInitializationMode()
    else:
        fmu.enterInitializationMode(tolerance=relative_tolerance, startTime=start_time)
        input.apply(time)
        fmu.exitInitializationMode()

    recorder = Recorder(fmu=fmu, modelDescription=model_description, variableNames=output, interval=output_interval)

    n_steps = 0

    # simulation loop
    while time < stop_time:

        if timeout is not None and (current_time() - sim_start) > timeout:
            break

        recorder.sample(time)

        input.apply(time)

        fmu.doStep(currentCommunicationPoint=time, communicationStepSize=output_interval)

        if step_finished is not None and not step_finished(time, recorder):
            break

        n_steps += 1
        time = n_steps * output_interval

    recorder.sample(time, force=True)

    fmu.terminate()

    return recorder.result()



import os
from ctypes import *
from .. import platform, sharedLibraryExtension

library_dir, _ = os.path.split(__file__)

logging = cdll.LoadLibrary(os.path.join(library_dir, platform, 'logging' + sharedLibraryExtension))

""" Adds a native proxy function that formats the message string using the variadic arguments that cannot be 
passed to Python with ctypes """
addLoggerProxy = getattr(logging, 'addLoggerProxy')
addLoggerProxy.argtypes = [c_void_p]  # pointer to fmi1CallbackFunctions or fmi2CallbackFunctions
addLoggerProxy.restype = None



from ctypes import cdll
import os
from fmpy import sharedLibraryExtension, platform_tuple

library_dir, _ = os.path.split(__file__)

# load SUNDIALS shared libraries
sundials_nvecserial     = cdll.LoadLibrary(os.path.join(library_dir, platform_tuple, 'sundials_nvecserial'     + sharedLibraryExtension))
sundials_sunmatrixdense = cdll.LoadLibrary(os.path.join(library_dir, platform_tuple, 'sundials_sunmatrixdense' + sharedLibraryExtension))
sundials_sunlinsoldense = cdll.LoadLibrary(os.path.join(library_dir, platform_tuple, 'sundials_sunlinsoldense' + sharedLibraryExtension))
sundials_cvode          = cdll.LoadLibrary(os.path.join(library_dir, platform_tuple, 'sundials_cvode'          + sharedLibraryExtension))





def add_cswrapper(filename, outfilename=None):

    from fmpy import read_model_description, extract, sharedLibraryExtension, platform, __version__
    from lxml import etree
    import os
    from shutil import copyfile, rmtree

    if outfilename is None:
        outfilename = filename

    model_description = read_model_description(filename)

    if model_description.fmiVersion != '2.0':
        raise Exception("%s is not an FMI 2.0 FMU." % filename)

    if model_description.modelExchange is None:
        raise Exception("%s does not support Model Exchange." % filename)

    unzipdir = extract(filename)

    xml = os.path.join(unzipdir, 'modelDescription.xml')

    tree = etree.parse(xml)

    root = tree.getroot()

    # update description
    generation_tool = root.attrib.get('generationTool', 'Unknown') + " with FMPy %s Co-Simulation wrapper" % __version__
    root.attrib['generationTool'] = generation_tool

    # remove any existing <CoSimulation> element
    for e in root.findall('CoSimulation'):
        root.remove(e)

    for i, child in enumerate(root):
        if child.tag == 'ModelExchange':
            break

    model_identifier = '%s_%s_%s' % (model_description.modelExchange.modelIdentifier,
                                     model_description.numberOfContinuousStates,
                                     model_description.numberOfEventIndicators)

    e = etree.Element("CoSimulation")
    e.attrib['modelIdentifier'] = model_identifier
    root.insert(i + 1, e)

    tree.write(xml, pretty_print=True, encoding='utf-8')

    shared_library = os.path.join(os.path.dirname(__file__), 'cswrapper' + sharedLibraryExtension)
    license_file = os.path.join(os.path.dirname(__file__), 'license.txt')

    licenses_dir = os.path.join(unzipdir, 'documentation', 'licenses')

    if not os.path.isdir(licenses_dir):
        os.mkdir(licenses_dir)

    copyfile(src=shared_library, dst=os.path.join(unzipdir, 'binaries', platform, model_identifier + sharedLibraryExtension))
    copyfile(license_file, os.path.join(unzipdir, 'documentation', 'licenses', 'fmpy-cswrapper.txt'))

    create_zip_archive(outfilename, unzipdir)

    rmtree(unzipdir, ignore_errors=True)


def create_zip_archive(filename, source_dir):

    import zipfile
    import os

    with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as zf:
        base_path = os.path.normpath(source_dir)
        for dirpath, dirnames, filenames in os.walk(source_dir):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, os.path.relpath(path, base_path))
            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, os.path.relpath(path, base_path))


if __name__ is '__main__':

    add_cswrapper('/Users/tors10/Development/Reference-FMUs/build/dist/Dahlquist.fmu',
                  '/Users/tors10/Development/Reference-FMUs/build/dist/DahlquistCS1.fmu')



from .libraries import sundials_sunlinsoldense
from .sundials_linearsolver import *
from .sundials_matrix import *
from .sundials_nvector import *

# /*
#  * -----------------------------------------------------------------
#  * Programmer(s): Daniel Reynolds, Ashley Crawford @ SMU
#  * -----------------------------------------------------------------
#  * SUNDIALS Copyright Start
#  * Copyright (c) 2002-2019, Lawrence Livermore National Security
#  * and Southern Methodist University.
#  * All rights reserved.
#  *
#  * See the top-level LICENSE and NOTICE files for details.
#  *
#  * SPDX-License-Identifier: BSD-3-Clause
#  * SUNDIALS Copyright End
#  * -----------------------------------------------------------------
#  * This is the header file for the dense implementation of the
#  * SUNLINSOL module, SUNLINSOL_DENSE.
#  *
#  * Notes:
#  *   - The definition of the generic SUNLinearSolver structure can
#  *     be found in the header file sundials_linearsolver.h.
#  *   - The definition of the type 'realtype' can be found in the
#  *     header file sundials_types.h, and it may be changed (at the
#  *     configuration stage) according to the user's needs.
#  *     The sundials_types.h file also contains the definition
#  *     for the type 'booleantype' and 'indextype'.
#  * -----------------------------------------------------------------
#  */
#
# #ifndef _SUNLINSOL_DENSE_H
# #define _SUNLINSOL_DENSE_H
#
# #include <sundials/sundials_linearsolver.h>
# #include <sundials/sundials_matrix.h>
# #include <sundials/sundials_nvector.h>
# #include <sundials/sundials_dense.h>
# #include <sunmatrix/sunmatrix_dense.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
# /* ----------------------------------------
#  * Dense Implementation of SUNLinearSolver
#  * ---------------------------------------- */
#
# struct _SUNLinearSolverContent_Dense {
#   sunindextype N;
#   sunindextype *pivots;
#   sunindextype last_flag;
# };
#
# typedef struct _SUNLinearSolverContent_Dense *SUNLinearSolverContent_Dense;
#
# /* ----------------------------------------
#  * Exported Functions for SUNLINSOL_DENSE
#  * ---------------------------------------- */
#
# SUNDIALS_EXPORT SUNLinearSolver SUNLinSol_Dense(N_Vector y, SUNMatrix A);
SUNLinSol_Dense = getattr(sundials_sunlinsoldense, 'SUNLinSol_Dense')
SUNLinSol_Dense.argtypes = [N_Vector, SUNMatrix]
SUNLinSol_Dense.restype = SUNLinearSolver
#
# /* deprecated */
# SUNDIALS_EXPORT SUNLinearSolver SUNDenseLinearSolver(N_Vector y,
#                                                      SUNMatrix A);
#
# SUNDIALS_EXPORT SUNLinearSolver_Type SUNLinSolGetType_Dense(SUNLinearSolver S);
# SUNDIALS_EXPORT SUNLinearSolver_ID SUNLinSolGetID_Dense(SUNLinearSolver S);
# SUNDIALS_EXPORT int SUNLinSolInitialize_Dense(SUNLinearSolver S);
# SUNDIALS_EXPORT int SUNLinSolSetup_Dense(SUNLinearSolver S, SUNMatrix A);
# SUNDIALS_EXPORT int SUNLinSolSolve_Dense(SUNLinearSolver S, SUNMatrix A,
#                                          N_Vector x, N_Vector b, realtype tol);
# SUNDIALS_EXPORT sunindextype SUNLinSolLastFlag_Dense(SUNLinearSolver S);
# SUNDIALS_EXPORT int SUNLinSolSpace_Dense(SUNLinearSolver S,
#                                          long int *lenrwLS,
#                                          long int *leniwLS);
# SUNDIALS_EXPORT int SUNLinSolFree_Dense(SUNLinearSolver S);
#
# #ifdef __cplusplus
# }
# #endif
#
# #endif



from ctypes import *

# /* -----------------------------------------------------------------
#  * Programmer(s): Daniel Reynolds @ SMU
#  *                David Gardner, Carol Woodward, Slaven Peles,
#  *                Cody Balos @ LLNL
#  * -----------------------------------------------------------------
#  * SUNDIALS Copyright Start
#  * Copyright (c) 2002-2019, Lawrence Livermore National Security
#  * and Southern Methodist University.
#  * All rights reserved.
#  *
#  * See the top-level LICENSE and NOTICE files for details.
#  *
#  * SPDX-License-Identifier: BSD-3-Clause
#  * SUNDIALS Copyright End
#  * -----------------------------------------------------------------
#  * This is the header file for a generic matrix package.
#  * It defines the SUNMatrix structure (_generic_SUNMatrix) which
#  * contains the following fields:
#  *   - an implementation-dependent 'content' field which contains
#  *     the description and actual data of the matrix
#  *   - an 'ops' filed which contains a structure listing operations
#  *     acting on such matrices
#  * -----------------------------------------------------------------
#  * This header file contains:
#  *   - enumeration constants for all SUNDIALS-defined matrix types,
#  *     as well as a generic type for user-supplied matrix types,
#  *   - type declarations for the _generic_SUNMatrix and
#  *     _generic_SUNMatrix_Ops structures, as well as references to
#  *     pointers to such structures (SUNMatrix), and
#  *   - prototypes for the matrix functions which operate on
#  *     SUNMatrix objects.
#  * -----------------------------------------------------------------
#  * At a minimum, a particular implementation of a SUNMatrix must
#  * do the following:
#  *   - specify the 'content' field of SUNMatrix,
#  *   - implement the operations on those SUNMatrix objects,
#  *   - provide a constructor routine for new SUNMatrix objects
#  *
#  * Additionally, a SUNMatrix implementation may provide the following:
#  *   - macros to access the underlying SUNMatrix data
#  *   - a routine to print the content of a SUNMatrix
#  * -----------------------------------------------------------------*/
#
# #ifndef _SUNMATRIX_H
# #define _SUNMATRIX_H
#
# #include <sundials/sundials_types.h>
# #include <sundials/sundials_nvector.h>
#
# #ifdef __cplusplus  /* wrapper to enable C++ usage */
# extern "C" {
# #endif
#
#
# /* -----------------------------------------------------------------
#  * Implemented SUNMatrix types
#  * ----------------------------------------------------------------- */
#
# typedef enum {
#   SUNMATRIX_DENSE,
#   SUNMATRIX_BAND,
#   SUNMATRIX_SPARSE,
#   SUNMATRIX_SLUNRLOC,
#   SUNMATRIX_CUSTOM
# } SUNMatrix_ID;
#
#
# /* -----------------------------------------------------------------
#  * Generic definition of SUNMatrix
#  * ----------------------------------------------------------------- */
#
# /* Forward reference for pointer to SUNMatrix_Ops object */
# typedef _SUNDIALS_STRUCT_ _generic_SUNMatrix_Ops *SUNMatrix_Ops;
#
# /* Forward reference for pointer to SUNMatrix object */
# typedef _SUNDIALS_STRUCT_ _generic_SUNMatrix *SUNMatrix;
SUNMatrix = c_void_p
#
# /* Structure containing function pointers to matrix operations  */
# struct _generic_SUNMatrix_Ops {
#   SUNMatrix_ID (*getid)(SUNMatrix);
#   SUNMatrix    (*clone)(SUNMatrix);
#   void         (*destroy)(SUNMatrix);
#   int          (*zero)(SUNMatrix);
#   int          (*copy)(SUNMatrix, SUNMatrix);
#   int          (*scaleadd)(realtype, SUNMatrix, SUNMatrix);
#   int          (*scaleaddi)(realtype, SUNMatrix);
#   int          (*matvecsetup)(SUNMatrix);
#   int          (*matvec)(SUNMatrix, N_Vector, N_Vector);
#   int          (*space)(SUNMatrix, long int*, long int*);
# };
#
# /* A matrix is a structure with an implementation-dependent
#    'content' field, and a pointer to a structure of matrix
#    operations corresponding to that implementation.  */
# struct _generic_SUNMatrix {
#   void *content;
#   SUNMatrix_Ops ops;
# };
#
#
# /* -----------------------------------------------------------------
#  * Functions exported by SUNMatrix module
#  * ----------------------------------------------------------------- */
#
# SUNDIALS_EXPORT SUNMatrix SUNMatNewEmpty();
# SUNDIALS_EXPORT void SUNMatFreeEmpty(SUNMatrix A);
# SUNDIALS_EXPORT int SUNMatCopyOps(SUNMatrix A, SUNMatrix B);
# SUNDIALS_EXPORT SUNMatrix_ID SUNMatGetID(SUNMatrix A);
# SUNDIALS_EXPORT SUNMatrix SUNMatClone(SUNMatrix A);
# SUNDIALS_EXPORT void SUNMatDestroy(SUNMatrix A);
# SUNDIALS_EXPORT int SUNMatZero(SUNMatrix A);
# SUNDIALS_EXPORT int SUNMatCopy(SUNMatrix A, SUNMatrix B);
# SUNDIALS_EXPORT int SUNMatScaleAdd(realtype c, SUNMatrix A, SUNMatrix B);
# SUNDIALS_EXPORT int SUNMatScaleAddI(realtype c, SUNMatrix A);
# SUNDIALS_EXPORT int SUNMatMatvecSetup(SUNMatrix A);
# SUNDIALS_EXPORT int SUNMatMatvec(SUNMatrix A, N_Vector x, N_Vector y);
# SUNDIALS_EXPORT int SUNMatSpace(SUNMatrix A, long int *lenrw, long int *leniw);
#
# /*
#  * -----------------------------------------------------------------
#  * IV. SUNMatrix error codes
#  * ---------------------------------------------------------------
#  */
#
# #define SUNMAT_SUCCESS                      0  /* function successfull          */
# #define SUNMAT_ILL_INPUT                 -701  /* illegal function input        */
# #define SUNMAT_MEM_FAIL                  -702  /* failed memory access/alloc    */
# #define SUNMAT_OPERATION_FAIL            -703  /* a SUNMatrix operation returned nonzero */
# #define SUNMAT_MATVEC_SETUP_REQUIRED     -704  /* the SUNMatMatvecSetup routine needs to be called */
#
# #ifdef __cplusplus
# }
# #endif
# #endif
